import pygame

from teon.functions import key_pressed,scale_def,_repair_vec2
from teon.entity import Entity
from teon.other import Vec2

class Controller2D(Entity):
    _instances = []
    def __init__(self,**kwargs):
        super().__init__(**kwargs)

        self.space = False

        self._position = Vec2((self.rect.centerx,self.rect.centery))

        Controller2D._instances.append(self)
        self.tags.append("dt")

        self._occluder_rect = pygame.Rect(1,1,pygame.display.get_window_size()[0],pygame.display.get_window_size()[1])

class PlayerController2D(Controller2D):
    _instances = []
    def __init__(self,**kwargs):
        self.can_collide = kwargs.get("can_collide",True)
        super().__init__(**kwargs)
        self.controller = kwargs.get("controller","full topdown")
        self.gravity = kwargs.get("gravity",0)
        self.screen_bound = kwargs.get("screen_bound",False)
        
        self.speed = kwargs.get("speed",300)
        self.can_jump = kwargs.get("can_jump",False)
        self.jump_force = kwargs.get("jump_force",0)
        self.jump_number = kwargs.get("jump_number",1)
        self.velocity = 0
        self.jumps = self.jump_number
        self.is_ui = False

        self.direction = Vec2((0,0))

        self.dt = 0

        self.space = False

        self._position = Vec2((self.rect.centerx,self.rect.centery))

        PlayerController2D._instances.append(self)

        self.tags.append("dt")

        self._last_position = _repair_vec2(self.hitbox.center)

    def _input(self):
        if self.controller == "platformer":
            if key_pressed("left"):
                self.direction.x = -1
            elif key_pressed("right"):
                self.direction.x = 1
            else:
                self.direction.x = 0

            if self.can_jump:
                if key_pressed("space") and self.jumps > 0 and not self.space:
                    self.velocity = -self.jump_force
                    self.jumps -= 1
                    self.space = True
                elif self.jumps > 0 and not key_pressed("space"):
                    self.space = False

        if self.controller == "topdown":
            if key_pressed("left"):
                self.direction.x = -1
                self.direction.y = 0
            elif key_pressed("right"):
                self.direction.x = 1
                self.direction.y = 0

            elif key_pressed("up"):
                self.direction.y = -1
                self.direction.x = 0
            elif key_pressed("down"):
                self.direction.y = 1
                self.direction.x = 0
            else:
                self.direction.x = 0
                self.direction.y = 0
        
        if self.controller == "full topdown":
            input_vector = Vec2()
            if key_pressed("up"):
                input_vector.y -= 1
            if key_pressed("down"):
                input_vector.y += 1
            if key_pressed("left"):
                input_vector.x -= 1
            if key_pressed("right"):
                input_vector.x += 1
            self.direction = input_vector.normalize() if input_vector else input_vector

    def _move(self):
        if self.controller == "platformer":
            self.velocity += self.gravity * self.dt
            self.position.y += self.velocity * self.dt

        self.position.x += self.direction.x * self.speed * self.dt

        self.position.y += self.direction.y * self.speed * self.dt

        self.hitbox.update()

        self._collide()

    def _collide(self):
        if self.can_collide:
            active_level = Entity.level_editor.active_level
            if active_level and active_level.index == self.level_index:
                self.collision_sprites = active_level.entities
                for sprite in self.collision_sprites:
                    if sprite.hitbox.colliderect(self.hitbox) and sprite is not self and sprite.collidable:
                        last_rect = pygame.Rect(self._last_position.x,self._last_position.y,self.hitbox.width,self.hitbox.height)

                        horizontal_overlap = max(0, min(last_rect.right, sprite.hitbox.right) - max(last_rect.left, sprite.hitbox.left))
                        vertical_overlap = max(0, min(last_rect.bottom, sprite.hitbox.bottom) - max(last_rect.top, sprite.hitbox.top))


                        pos = self.hitbox.center
                        
                        if (horizontal_overlap,vertical_overlap) == (0,0):
                            horizontal_overlap = min(self.hitbox.right - sprite.hitbox.left, sprite.hitbox.right - self.hitbox.left)
                            vertical_overlap = min(self.hitbox.bottom - sprite.hitbox.top, sprite.hitbox.bottom - self.hitbox.top)
                            
                            if abs(horizontal_overlap) < abs(vertical_overlap):
                                if self.hitbox.centerx <= sprite.hitbox.centerx:
                                    self.hitbox.right = sprite.hitbox.left
                                else:
                                    self.hitbox.left = sprite.hitbox.right
                                self.position.x += (self.hitbox.centerx - pos[0]) / scale_def()

                            else:
                                if self.hitbox.centery <= sprite.hitbox.centery:
                                    self.hitbox.bottom = sprite.hitbox.top
                                    self.velocity = 0
                                    self.jumps = self.jump_number
                                else:
                                    self.hitbox.top = sprite.hitbox.bottom
                                    self.velocity = 0
                                self.position.y += (self.hitbox.centery - pos[1]) / scale_def()

                        elif horizontal_overlap <= vertical_overlap:
                            if self.hitbox.centerx <= sprite.hitbox.centerx:
                                self.hitbox.right = sprite.hitbox.left
                            else:
                                self.hitbox.left = sprite.hitbox.right
                            self.position.x += (self.hitbox.centerx - pos[0]) / scale_def()

                        else:
                            if self.hitbox.centery <= sprite.hitbox.centery:
                                self.hitbox.bottom = sprite.hitbox.top
                                self.velocity = 0
                                self.jumps = self.jump_number
                            else:
                                self.hitbox.top = sprite.hitbox.bottom
                                self.velocity = 0
                            self.position.y += (self.hitbox.centery - pos[1]) / scale_def()

        self._last_position = _repair_vec2(self.hitbox.topleft)
                         
    def _bound(self):
        if self.rect.x < 0:
            self.rect.x = 0
            self._position = self.rect.center
        if self.rect.y < 0:
            self.rect.y = 0
            self._position = self.rect.center
        if self.rect.right > pygame.display.get_window_size()[0]:
            self.rect.right = pygame.display.get_window_size()[0]
            self._position = self.rect.center
        if self.rect.bottom > pygame.display.get_window_size()[1]:
            self.rect.bottom = pygame.display.get_window_size()[1]
            self._position = self.rect.center
            self.velocity = 0
            self.jumps = self.jump_number

    def update(self):
        super().update()
        self._occlusion_rect.center = self.rect.center
        self._occluder_rect.center = self.rect.center
        self._update_colliders()
        self._input()
        self._move()
        if self.screen_bound:
            self._bound()